// --8<--8<--8<--8<--
//
// Copyright (C) 2006 Smithsonian Astrophysical Observatory
//
// This file is part of scatter
//
// scatter is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// scatter is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the 
//       Free Software Foundation, Inc. 
//       51 Franklin Street, Fifth Floor
//       Boston, MA  02110-1301, USA
//
// -->8-->8-->8-->8--

#include "ScatterDatabase.h"

ScatterDatabase::ScatterDatabase( 
  const string& scatterdb_filename, 
  const int surfno,
  const int shell,
  const double z0
 ) throw ( Exception ) {

  TraceFct tf( "ScatterDatabase::ScatterDatabase" );

  
  string scatterdb_filename_in;

  // interpolate scatterdb_filename for environmental variables
  {
    int error;
    char *temp = str_interp( scatterdb_filename.c_str( ), 1, NULL, NULL,
			     &error );
    if ( error )
      throw Exception( string( "error in interpolation of file name: " )
		       + scatterdb_filename );

    scatterdb_filename_in = temp;
    free( temp );
  }

  string mirror;
  string zoneid;
  string scatalg;
  string scatfile;

  RDB scatterdb( scatterdb_filename_in );
  scatterdb.mapData( "mirror",    &mirror   );
  scatterdb.mapData( "zone",      &zoneid   );
  scatterdb.mapData( "scat_alg",  &scatalg  );
  scatterdb.mapData( "scat_file", &scatfile );

  stringstream mirror_ss;
  mirror_ss << (1 == surfno ? 'p' : 'h' ) << shell;
  string requested_mirror = mirror_ss.str( );

  while ( scatterdb.read( ) ) {

    if ( requested_mirror == mirror ) {

      // interpolate scat_file for environmental variables
      int error;
      char *temp = str_interp( scatfile.c_str(), 1, NULL, NULL, &error );
      if ( error )
	throw Exception( string( "error in interpolation of file name: " )
			 + scatfile );
      string scatfile_in( temp );
      free( temp );

      if ( "EsaLinAsym" == scatalg ) {
	_zones.push_back( new MZN( zoneid, scatfile_in, scatterdb ) );
      }
      else if ( "EsaLinPowSym" == scatalg ) 
	_zones.push_back( new MZO( zoneid, scatfile_in, scatterdb ) );
      
      else 
	throw( Exception( string("unsupported scattering algorithm '") + scatalg + "' specified" ) );


    }

  }

  if ( ! _zones.size( ) ) 
    throw( Exception( string("mirror '") + requested_mirror + "' not found in scatter database:  " + scatterdb_filename_in ) );

  // now, sort the zones to ensure that they're in ascending order...
  sort( _zones.begin( ), _zones.end( ), mz_less() );

  //  check that zones don't overlap...
  _zones[0]->zmin( _zones[0]->zmin( ) + z0 );
  _zones[0]->zmax( _zones[0]->zmax( ) + z0 );

  for ( vector<MZ*>::iterator itr = _zones.begin( ) + 1; _zones.end( ) != itr; itr++ ) {
    (*itr)->zmin( (*itr)->zmin( ) + z0 );
    (*itr)->zmax( (*itr)->zmax( ) + z0 );
    
    if ( (*(itr-1))->zmin() > (*itr)->zmin() || (*(itr-1))->zmax() > (*itr)->zmax() ) {

      stringstream ss;
      ss << "zones " << (*(itr-1))->id( ) << " and " << (*itr)->id( ) << " overlap";
      throw( Exception( ss.str( ) ) );

    }
  }
}

ScatterDatabase::~ScatterDatabase( void ) {
  
  TraceFct tf( "ScatterDatabase::~ScatterDatabase" );

  for ( vector<MZ*>::iterator itr = _zones.begin( ); _zones.end( ) != itr; itr++ ) delete *itr;

}
/*
void ScatterDatabase::scatter( ParFile& parfile, Photon& photon ) throw ( Exception ) {

  TraceFct tf( "ScatterDatabase::scatter" );

  

}
*/
MZ *ScatterDatabase::findZone( const double z ) throw ( Exception ) {

  TraceFct tf( "ScatterDatabase::findZone" );


  /*
    binary search, ref NR, 2ed, p117. since zones are guaranteed
    to ascend, modified to remove bidirectional feature
  */
  int jl, ju;
  int j = 0;
  
  if ( 0 == _zones.size() )
    throw( Exception( "internal error: no zones?" ) );

  jl = -1;
  ju = _zones.size( );
  MZ *zone = NULL;

  while ( ju - jl > 1 ) {
    j = (ju + jl) >> 1;

    if ( (_zones[j]->zmin() < z) && (z < _zones[j]->zmax()) )
    {
      zone = _zones[j];
      break;
    }
    else if (z < _zones[j]->zmin())
      ju = j;
    else	  /* (z > _zones[j]->zmax()) */
      jl = j;
  }
  
  if ( (jl > -1) && (ju < _zones.size()) &&
       _zones[j]->zmin() <= z && z <= _zones[j]->zmax() )
    zone = _zones[j];
  
  return zone;

}


